----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    17:11:42 07/14/2014 
-- Design Name: 
-- Module Name:    RHA_TO_ZL_CONVERTER - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
--
----------------------------------------------------------------------------------
library IEEE;

use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;


entity RHA_TO_ZL_CONVERTER is
    Port (

		-- This we get from the RHA side: 
	   	DATAOUT_DATA_OUT			: in  STD_LOGIC_VECTOR(15 downto 0);
		DATAOUT_VALID_OUT			: in  STD_LOGIC;
		NewBlock						: in  STD_LOGIC;
		
		ReadRequest					: out STD_LOGIC;

		-- This goes into the Zarlink / FIFO direction: 
		FIFO_OUT_DATA   			: out STD_LOGIC_VECTOR(7 downto 0);
		FIFO_OUT_VALID   			: out STD_LOGIC;

		-- Parameters for the header
		ChannelMask     			: in  std_logic_vector(127 downto 0);
		ADCResolution   			: in  std_logic_vector(3 downto 0);
		SampleRate      			: in  std_logic_vector(7 downto 0);
		Filter          			: in  std_logic_vector(3 downto 0);
		DeviceID        			: in  std_logic_vector(7 downto 0);		
		
		-- Bitmask for the data selection from the 16 bit samples.
		BITMASK_SAMPLE				: in  std_logic_vector(15 downto 0);		
		
		-- This is the signal from the fifo that it still has room
		-- for a complete block of 128 channels. 
		-- Calculation for the worst case:
		-- 128 * 16 bit samples = 2048
		-- 174 bits for the largest possible header 
		-- (it is 176 but 2 bits are for signaling whether it is 
		-- the beginning of the block or not.)
		-- = 2222 bit
		-- One Zarlink transaction covers 14*8 bits = 112 bit
		-- However only 111 are usable for out payload
		-- the other bit we use for signaling whether 
		-- this is the beginning of a new 128 channel block.
		-- Thus we want the FIFO to have (at least)
		-- 280 x 8 bit (2240 bits)
		-- of free space for the next block to be on the safe side.
		ExportNext128ChannelBlock:in STD_LOGIC; 
		
		Clk 			        		: in  STD_LOGIC;
		-- This is the clock for the RHA machines 
		-- we need it for the domain crossing 
		Clk_SLOW		        		: in  STD_LOGIC;
      Reset 		            : in  STD_LOGIC	
	  );
end RHA_TO_ZL_CONVERTER;

architecture Behavioral of RHA_TO_ZL_CONVERTER is

	-- For the headers... 
	signal Header_Long_A   	: std_logic_vector(111 downto 0) := (others => '0');
	signal Header_Long_B   	: std_logic_vector( 63 downto 0) := (others => '0');
	signal Header_Short    	: std_logic_vector( 30 downto 0) := (others => '0');
	signal Timestamp       	: unsigned(15 downto 0) := (others => '0');
	constant PacketHeader_0 : std_logic_vector(3 downto 0) := "0000";
	constant PacketHeader_1 : std_logic_vector(3 downto 0) := "0001";

	signal FIFO_OUT_DATA_INTERNAL	: STD_LOGIC_VECTOR(7 downto 0) := (others => '0');
	signal FIFO_OUT_VALID_INTERNAL: STD_LOGIC := '0';

	signal BITMASK_SAMPLE_COPY	: std_logic_vector(15 downto 0) := (others => '0');
	signal DATA_COPY				: std_logic_vector(15 downto 0) := (others => '0');
	signal BITS_TO_PROCESS		: std_logic_vector(15 downto 0) := (others => '0');
	
	-- Needs to count up to 112
	signal COUNTER_BITS_PROCESSED : unsigned(6 downto 0) := (others => '0');
	-- Needs to count up to 8
	signal COUNTER_BITS_OF_A_BYTE	: unsigned(3 downto 0) := (others => '0');
	
	-- We use the long header only once 
	signal FIRST_HEADER_USED	: STD_LOGIC := '0';

	signal NewBlock_LAST 		: STD_LOGIC := '0';
	signal ReadRequest_INTERNAL : STD_LOGIC := '0';
	signal ReadRequest_INTERNAL_LAST : STD_LOGIC := '0';
	signal ReadRequest_INTERNAL_LAST2 : STD_LOGIC := '0';
	signal ReadRequest_INTERNAL_LAST3 : STD_LOGIC := '0';
	signal ReadRequest_INTERNAL_LAST4 : STD_LOGIC := '0';

	signal DATAOUT_DATA_OUT_BUFFER		: STD_LOGIC_VECTOR(15 downto 0) := (others => '0');
	signal DATAOUT_VALID_OUT_BUFFER		: STD_LOGIC := '0';
	signal DATA_USED							: STD_LOGIC := '0';
	
	type STATE_Type is (
		-- Wait until there is enougth space left and a new block begins
		SM_IDLE,
		-- Header_Long_A(7 downto 0)
		SM_HEADER_LONG_A_01,
		-- Header_Long_A(15 downto 8)
		SM_HEADER_LONG_A_02,
		-- Header_Long_A(23 downto 16)
		SM_HEADER_LONG_A_03,
		-- Header_Long_A(31 downto 24)
		SM_HEADER_LONG_A_04,
		-- Header_Long_A(39 downto 32)
		SM_HEADER_LONG_A_05,
		-- Header_Long_A(47 downto 40)
		SM_HEADER_LONG_A_06,
		-- Header_Long_A(55 downto 48)
		SM_HEADER_LONG_A_07,
		-- Header_Long_A(63 downto 56)
		SM_HEADER_LONG_A_08,
		-- Header_Long_A(71 downto 64)
		SM_HEADER_LONG_A_09,
		-- Header_Long_A(79 downto 72)
		SM_HEADER_LONG_A_10,
		-- Header_Long_A(87 downto 80)
		SM_HEADER_LONG_A_11,
		-- Header_Long_A(95 downto 88)
		SM_HEADER_LONG_A_12,
		-- Header_Long_A(103 downto 96)
		SM_HEADER_LONG_A_13,
		-- Header_Long_A(111 downto 104)
		SM_HEADER_LONG_A_14,
		-- Header_Long_B(7 downto 0)
		SM_HEADER_LONG_B_01,
		-- Header_Long_B(15 downto 8)
		SM_HEADER_LONG_B_02,
		-- Header_Long_B(23 downto 16)
		SM_HEADER_LONG_B_03,
		-- Header_Long_B(31 downto 24)
		SM_HEADER_LONG_B_04,
		-- Header_Long_B(39 downto 32)
		SM_HEADER_LONG_B_05,
		-- Header_Long_B(47 downto 40)
		SM_HEADER_LONG_B_06,
		-- Header_Long_B(55 downto 48)
		SM_HEADER_LONG_B_07,
		-- Header_Long_B(63 downto 56)	
		SM_HEADER_LONG_B_08,		
		SM_HEADER_LONG_B_END,
		-- Header_Short(7 downto 0)
		SM_HEADER_SHORT_01,
		-- Header_Short(15 downto 8)
		SM_HEADER_SHORT_02,
		-- Header_Short(23 downto 16)
		SM_HEADER_SHORT_03,
		-- Header_Short(30 downto 24)
		SM_HEADER_SHORT_UNFINISHED,
		SM_COPY_DATA_BITS_REQUEST,
		SM_COPY_DATA_BITS_START,
		SM_COPY_DATA_BITS_NEWLINE,
		SM_CLEAN_UP
	);	
	signal SPIMACHINE_STATE : STATE_TYPE;
	signal NewBlock_Off : STD_LOGIC := '0';
	signal NewBlock_INTERNAL : STD_LOGIC := '0';
	
	signal ChannelMask_Resorted : std_logic_vector(127 downto 0) := (others => '0');
	
begin
	-- Export internal and buffered signals 
	FIFO_OUT_DATA <= FIFO_OUT_DATA_INTERNAL;
	FIFO_OUT_VALID <= FIFO_OUT_VALID_INTERNAL;
	
	-- Notes: 
	-- 1.) We need to build 112bit long strings 
	--     every string starts with a '0' except the first one of a 128 channel block
	-- 2.) The header long is only used one time!

	ChannelMask_Resorted <= 
	  ChannelMask(127) & ChannelMask(111) & ChannelMask(95) & ChannelMask(79)
	& ChannelMask(63) & ChannelMask(47) & ChannelMask(31) & ChannelMask(15)
	& ChannelMask(126) & ChannelMask(110) & ChannelMask(94) & ChannelMask(78)
	& ChannelMask(62) & ChannelMask(46) & ChannelMask(30) & ChannelMask(14)
	& ChannelMask(125) & ChannelMask(109) & ChannelMask(93) & ChannelMask(77)
	& ChannelMask(61) & ChannelMask(45) & ChannelMask(29) & ChannelMask(13)
	& ChannelMask(124) & ChannelMask(108) & ChannelMask(92) & ChannelMask(76)
	& ChannelMask(60) & ChannelMask(44) & ChannelMask(28) & ChannelMask(12)
	& ChannelMask(123) & ChannelMask(107) & ChannelMask(91) & ChannelMask(75)
	& ChannelMask(59) & ChannelMask(43) & ChannelMask(27) & ChannelMask(11)
	& ChannelMask(122) & ChannelMask(106) & ChannelMask(90) & ChannelMask(74)
	& ChannelMask(58) & ChannelMask(42) & ChannelMask(26) & ChannelMask(10)
	& ChannelMask(121) & ChannelMask(105) & ChannelMask(89) & ChannelMask(73)
	& ChannelMask(57) & ChannelMask(41) & ChannelMask(25) & ChannelMask(9)
	& ChannelMask(120) & ChannelMask(104) & ChannelMask(88) & ChannelMask(72)
	& ChannelMask(56) & ChannelMask(40) & ChannelMask(24) & ChannelMask(8)
	& ChannelMask(119) & ChannelMask(103) & ChannelMask(87) & ChannelMask(71)
	& ChannelMask(55) & ChannelMask(39) & ChannelMask(23) & ChannelMask(7)
	& ChannelMask(118) & ChannelMask(102) & ChannelMask(86) & ChannelMask(70)
	& ChannelMask(54) & ChannelMask(38) & ChannelMask(22) & ChannelMask(6)
	& ChannelMask(117) & ChannelMask(101) & ChannelMask(85) & ChannelMask(69)
	& ChannelMask(53) & ChannelMask(37) & ChannelMask(21) & ChannelMask(5)
	& ChannelMask(116) & ChannelMask(100) & ChannelMask(84) & ChannelMask(68)
	& ChannelMask(52) & ChannelMask(36) & ChannelMask(20) & ChannelMask(4)
	& ChannelMask(115) & ChannelMask(99) & ChannelMask(83) & ChannelMask(67)
	& ChannelMask(51) & ChannelMask(35) & ChannelMask(19) & ChannelMask(3)
	& ChannelMask(114) & ChannelMask(98) & ChannelMask(82) & ChannelMask(66)
	& ChannelMask(50) & ChannelMask(34) & ChannelMask(18) & ChannelMask(2)
	& ChannelMask(113) & ChannelMask(97) & ChannelMask(81) & ChannelMask(65)
	& ChannelMask(49) & ChannelMask(33) & ChannelMask(17) & ChannelMask(1)
	& ChannelMask(112) & ChannelMask(96) & ChannelMask(80) & ChannelMask(64)
	& ChannelMask(48) & ChannelMask(32) & ChannelMask(16) & ChannelMask(0);

	-- Here we compile the headers 
	Header_Long_A(0) <= '1'; -- Begin of a data set which contains all the selected electrodes
	Header_Long_A(4 downto 1) <= PacketHeader_0; -- Marks the type of the header (long)
	Header_Long_A(12 downto 5) <= DeviceID; -- ID of the ASIC 
	Header_Long_A(111 downto 13) <= ChannelMask_Resorted(98 downto 0); -- first part of the channel mask

	Header_Long_B(0) <= '0'; -- Not the begin of a data set => '0'!!!
	Header_Long_B(29 downto 1) <= ChannelMask_Resorted(127 downto 99);
	Header_Long_B(33 downto 30) <= Filter; -- std_logic_vector(3 downto 0)
	Header_Long_B(49 downto 34) <= std_logic_vector(Timestamp); -- unsigned(15 downto 0)
	Header_Long_B(57 downto 50) <= SampleRate; -- std_logic_vector(7 downto 0)
	Header_Long_B(61 downto 58) <= ADCResolution; -- std_logic_vector(3 downto 0)
	Header_Long_B(62) <= '0'; -- unused/control bit
	Header_Long_B(63) <= '0'; -- unused/control bit

	Header_Short(0) <= '1'; -- Begin of a data set which contains all the selected electrodes
	Header_Short(4 downto 1) <= PacketHeader_1; -- Marks the type of the header (short)
	Header_Short(12 downto 5) <= DeviceID; -- ID of the ASIC 
	Header_Short(28 downto 13) <= std_logic_vector(Timestamp); -- unsigned(15 downto 0)
	Header_Short(29) <= '0'; -- unused/control bit
	Header_Short(30) <= '0'; -- unused/control bit



	ReadRequest <= ReadRequest_INTERNAL;
	-- We need to decouple the data request from the rest 
	DATA_IMPORTING_PROCESS : process (Clk,Reset)
	begin
		if Reset = '1' then
			ReadRequest_INTERNAL <= '0';
			DATAOUT_DATA_OUT_BUFFER <= (others => '0');
			DATAOUT_VALID_OUT_BUFFER <= '0';
			ReadRequest_INTERNAL_LAST <= '0';
			ReadRequest_INTERNAL_LAST2 <= '0';
			ReadRequest_INTERNAL_LAST3 <= '0';			
			ReadRequest_INTERNAL_LAST4 <= '0';	
		elsif rising_edge(Clk) then

			ReadRequest_INTERNAL <= '0';
			DATAOUT_DATA_OUT_BUFFER <= DATAOUT_DATA_OUT_BUFFER;
			DATAOUT_VALID_OUT_BUFFER <= DATAOUT_VALID_OUT_BUFFER;

			ReadRequest_INTERNAL_LAST <= ReadRequest_INTERNAL;
			ReadRequest_INTERNAL_LAST2 <= ReadRequest_INTERNAL_LAST;
			ReadRequest_INTERNAL_LAST3 <= ReadRequest_INTERNAL_LAST2;
			ReadRequest_INTERNAL_LAST4 <= ReadRequest_INTERNAL_LAST3;
			
			if DATA_USED = '1' then
				DATAOUT_DATA_OUT_BUFFER <= (others => '0');
				DATAOUT_VALID_OUT_BUFFER <= '0';
				ReadRequest_INTERNAL <= '1';
			elsif (ReadRequest_INTERNAL = '1') OR (ReadRequest_INTERNAL_LAST = '1') then
				null;
			elsif DATAOUT_VALID_OUT = '1' then
				DATAOUT_DATA_OUT_BUFFER <= DATAOUT_DATA_OUT;
				DATAOUT_VALID_OUT_BUFFER <= '1';
			else
				null;
			end if;
	

		end if;
	end process DATA_IMPORTING_PROCESS;	

	
	-- We check when a new block arrives and count the timestamp one up. 
	TIME_STAMP_PROCESS : process (Clk,Reset)
	begin
		if Reset = '1' then
			Timestamp <= (others => '0');
			
		elsif rising_edge(Clk) then
			NewBlock_LAST <= NewBlock;
			Timestamp <= Timestamp;
			
			if (NewBlock = '1') AND (NewBlock_LAST = '0') then
				Timestamp <= Timestamp + 1;
			else
				null;
			end if;
			
		end if;
	end process TIME_STAMP_PROCESS;	

	
	FSM_BLOCK : process (Clk,Reset)
	begin
		if Reset = '1' then
			FIRST_HEADER_USED <= '0';
			FIFO_OUT_DATA_INTERNAL <= (others => '0');
			FIFO_OUT_VALID_INTERNAL <= '0';
			COUNTER_BITS_PROCESSED <= (others => '0');
			BITMASK_SAMPLE_COPY <= (others => '0');
			DATA_COPY <= (others => '0');
			BITS_TO_PROCESS <= (others => '0');
			COUNTER_BITS_OF_A_BYTE <= (others => '0');
			NewBlock_Off <= '0';
			NewBlock_INTERNAL <= '0';
			DATA_USED <= '0';
			SPIMACHINE_STATE <= SM_IDLE;

		elsif rising_edge(Clk) then
			FIRST_HEADER_USED <= FIRST_HEADER_USED;
			FIFO_OUT_DATA_INTERNAL <= FIFO_OUT_DATA_INTERNAL;
			FIFO_OUT_VALID_INTERNAL <= '0';
			COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED;
			BITMASK_SAMPLE_COPY <= BITMASK_SAMPLE_COPY;
			DATA_COPY <= DATA_COPY;
			BITS_TO_PROCESS <= BITS_TO_PROCESS;

			COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE;
			NewBlock_Off <= '0';
			DATA_USED <= '0';
			
			-- Failsafe !
			SPIMACHINE_STATE <= SM_IDLE;
			-- AND (ExportNext128ChannelBlock = '1')
			if (NewBlock = '1') AND (NewBlock_LAST = '0') AND (ExportNext128ChannelBlock = '1') then
				NewBlock_INTERNAL <= '1';
			elsif NewBlock_Off = '1' then 
				NewBlock_INTERNAL <= '0';
			else 
				NewBlock_INTERNAL <= NewBlock_INTERNAL;
			end if;

			case SPIMACHINE_STATE is
			
				-- Wait until something has to be done. Now we decide which header we need to use... 
				-- Use the long one only once!
				when SM_IDLE =>
					if (NewBlock_INTERNAL = '1') AND (ExportNext128ChannelBlock = '1') AND (FIRST_HEADER_USED = '0') then
						SPIMACHINE_STATE <= SM_HEADER_LONG_A_01;
						NewBlock_Off <= '1';
					elsif (NewBlock_INTERNAL = '1') AND (ExportNext128ChannelBlock = '1') AND (FIRST_HEADER_USED = '1') then						
						SPIMACHINE_STATE <= SM_HEADER_SHORT_01;
						NewBlock_Off <= '1';
					else
						SPIMACHINE_STATE <= SM_IDLE;
					end if;

					-- We need to wait for the next new block or new command 
					if (DATAOUT_VALID_OUT_BUFFER = '1') AND (NewBlock_INTERNAL = '0') then
						DATA_USED <= '1';
					-- We need to discard the newblock because we have no space in the fifo left 	
					elsif (DATAOUT_VALID_OUT_BUFFER = '1') AND (ExportNext128ChannelBlock = '0') then 
						DATA_USED <= '1';
					else
						null;
					end if;

----------------------------------

				when SM_HEADER_LONG_A_01 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_02;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(7 downto 0);
					FIFO_OUT_VALID_INTERNAL <= '1';
					-- Use the long header only once!
					FIRST_HEADER_USED <= '1';

				when SM_HEADER_LONG_A_02 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_03;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(15 downto 8);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_03 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_04;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(23 downto 16);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_04 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_05;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(31 downto 24);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_05 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_06;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(39 downto 32);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_06 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_07;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(47 downto 40);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_07 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_08;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(55 downto 48);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_08 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_09;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(63 downto 56);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_09 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_10;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(71 downto 64);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_10 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_11;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(79 downto 72);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_11 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_12;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(87 downto 80);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_12 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_13;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(95 downto 88);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_13 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_A_14;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(103 downto 96);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_A_14 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_01;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_A(111 downto 104);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_01 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_02;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(7 downto 0);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_02 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_03;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(15 downto 8);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_03 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_04;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(23 downto 16);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_04 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_05;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(31 downto 24);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_05 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_06;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(39 downto 32);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_06 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_07;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(47 downto 40);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_07 =>
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_08;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(55 downto 48);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_08 =>		
					SPIMACHINE_STATE <= SM_HEADER_LONG_B_END;
					FIFO_OUT_DATA_INTERNAL <= Header_Long_B(63 downto 56);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_LONG_B_END =>		
					SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
					FIFO_OUT_DATA_INTERNAL <= (others => '0');
					FIFO_OUT_VALID_INTERNAL <= '0';
					COUNTER_BITS_PROCESSED <= to_unsigned(64,7);
					COUNTER_BITS_OF_A_BYTE <= to_unsigned(0,4);

----------------------------------

				when SM_HEADER_SHORT_01 =>
					SPIMACHINE_STATE <= SM_HEADER_SHORT_02;
					FIFO_OUT_DATA_INTERNAL <= Header_Short(7 downto 0);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_SHORT_02 =>
					SPIMACHINE_STATE <= SM_HEADER_SHORT_03;
					FIFO_OUT_DATA_INTERNAL <= Header_Short(15 downto 8);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_SHORT_03 =>
					SPIMACHINE_STATE <= SM_HEADER_SHORT_UNFINISHED;
					FIFO_OUT_DATA_INTERNAL <= Header_Short(23 downto 16);
					FIFO_OUT_VALID_INTERNAL <= '1';

				when SM_HEADER_SHORT_UNFINISHED =>
					SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
					FIFO_OUT_DATA_INTERNAL <= Header_Short(30 downto 24) & "0";
					COUNTER_BITS_PROCESSED <= to_unsigned(31,7);
					COUNTER_BITS_OF_A_BYTE <= to_unsigned(7,4);

----------------------------------

				-- For the data: 	
					
				-- Here we request and wait for the next sample 
				when SM_COPY_DATA_BITS_REQUEST => 

					-- If we get during getting the new sample,
					-- a newblock signal or a new command then we have to close 
					-- the old block 

					if NewBlock_INTERNAL = '1' then
						SPIMACHINE_STATE <= SM_CLEAN_UP;
						NewBlock_Off <= '1';
					-- New data discovered. Do something with it. 
					elsif DATAOUT_VALID_OUT_BUFFER = '1' then 
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
						BITMASK_SAMPLE_COPY <= BITMASK_SAMPLE;
						DATA_COPY <= DATAOUT_DATA_OUT_BUFFER;
						BITS_TO_PROCESS <= (others => '1');
						DATA_USED <= '1';
					-- Wait for new data in this state 
					else
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
					end if;

				when SM_COPY_DATA_BITS_START =>
					-- All the data was used, we need to request new data 
					if BITS_TO_PROCESS(0) = '0' then
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_REQUEST;
					-- The Zarlink string is full and a new one needs to be created
					-- export the finished string to the fifo 
					elsif COUNTER_BITS_OF_A_BYTE = to_unsigned(8,4) then
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_NEWLINE;
						FIFO_OUT_VALID_INTERNAL <= '1';						
					-- We found a bit to process
					elsif (BITS_TO_PROCESS(0) = '1') AND (BITMASK_SAMPLE_COPY(0) = '1') then
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
						FIFO_OUT_DATA_INTERNAL <= DATA_COPY(0) & FIFO_OUT_DATA_INTERNAL(7 downto 1);
						COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 1;
						COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 1;
						BITMASK_SAMPLE_COPY <= "0" & BITMASK_SAMPLE_COPY(15 downto 1);
						DATA_COPY <= "0" & DATA_COPY(15 downto 1);
						BITS_TO_PROCESS <= "0" & BITS_TO_PROCESS(15 downto 1);
						-- This bit shouldn't be processed
					else
						SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
						BITMASK_SAMPLE_COPY <= "0" & BITMASK_SAMPLE_COPY(15 downto 1);
						DATA_COPY <= "0" & DATA_COPY(15 downto 1);
						BITS_TO_PROCESS <= "0" & BITS_TO_PROCESS(15 downto 1);
					end if;

				-- Here we begin a new Zarlink string. 
				when SM_COPY_DATA_BITS_NEWLINE =>
					SPIMACHINE_STATE <= SM_COPY_DATA_BITS_START;
					if COUNTER_BITS_PROCESSED = to_unsigned(112,7) then 
						-- It can only start with one "0"
						-- because it is never a header 					
						FIFO_OUT_DATA_INTERNAL <= (others => '0');
						COUNTER_BITS_PROCESSED <= to_unsigned(1,7);
						COUNTER_BITS_OF_A_BYTE <= to_unsigned(1,4);
					else
						-- We only finished a 8 bit block... 
						FIFO_OUT_DATA_INTERNAL <= (others => '0');
						COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED;
						COUNTER_BITS_OF_A_BYTE <= to_unsigned(0,4);
					end if;
					
				-- Fill up the last 112 bits with zeros... 
				when SM_CLEAN_UP =>

					-- If everything is filled up with zeros then...  
					-- go here
					if (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext128ChannelBlock = '0') then
						SPIMACHINE_STATE <= SM_IDLE;
						FIFO_OUT_VALID_INTERNAL <= '1';
					-- or here 
					-- The next one will be data
					elsif (COUNTER_BITS_PROCESSED = to_unsigned(112,7)) AND (ExportNext128ChannelBlock = '1') then
						SPIMACHINE_STATE <= SM_HEADER_SHORT_01;
						FIFO_OUT_VALID_INTERNAL <= '1';
					elsif COUNTER_BITS_OF_A_BYTE = to_unsigned(8,4) then
						-- Send another 8 bits on their way 
						SPIMACHINE_STATE <= SM_CLEAN_UP;					
						FIFO_OUT_VALID_INTERNAL <= '1';
						COUNTER_BITS_OF_A_BYTE <= (others => '0');
					elsif (COUNTER_BITS_OF_A_BYTE = to_unsigned(0,4)) AND (FIFO_OUT_VALID_INTERNAL = '1') then
						SPIMACHINE_STATE <= SM_CLEAN_UP;					
						FIFO_OUT_DATA_INTERNAL <= (others => '0');
						COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 8;
						COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 8;
					else 
					-- if zeros are missing continue here and stay in this state 
						SPIMACHINE_STATE <= SM_CLEAN_UP;
						FIFO_OUT_DATA_INTERNAL <= "0" & FIFO_OUT_DATA_INTERNAL(7 downto 1);
						COUNTER_BITS_PROCESSED <= COUNTER_BITS_PROCESSED + 1;
						COUNTER_BITS_OF_A_BYTE <= COUNTER_BITS_OF_A_BYTE + 1;
					end if;				

				when others =>
					SPIMACHINE_STATE <= SM_IDLE;
			end case;		
		end if;
	end process FSM_BLOCK;	
	
end architecture Behavioral;

